/*

2clseq. Copyright Fahiem Bacchus 2000--2002

 A DPLL solver based on performing extensive reasoning with the binary
 clause subtheory. Pronounced "Two" "Clause" "Eq"

Version 2.0.

The software available at this site is copyright (c) 2000-2002 Fahiem
Bacchus. All right are reserved. Use of this software is permitted for
non-commercial research purposes, and it may be copied only for that
use. All copies must include this copyright message. This software and
any documentation and/or information supplied with it is distributed
on an as is basis. Fahiem Bacchus and the University of Toronto make
no warranties, express or implied, including but not limited to
implied warranties of merchantability and fitness for a particular
purpose, regarding the documentation, functions or performance of such
software, documentation and/or information.

@misc{Bacchus:2clseq,
	author  = { Fahiem Bacchus },
	year    = { 2002 },
	title   = { {2clseq} a {DPLL} solver employing extensive binary clause reasoning },
	note    = { Available from 
                    http://www.cs.toronto.edu/\~{}fbacchus/2clseq.html }
}

*/

#include "2clsEq.h"
#include "undo.h"
#include "resolve.h"
#include "dag.h"
#include "nogood.h"
//#include "binaryClauses.h"

/***********************************************************************
                                DPL
***********************************************************************/
/* DPL (Davis Putnam Longham) is the backtracking engine. 

DPL calls findNext to choose the variable to value next.
   
We also configure DPL to count the leaves in the search tree. This
corresponds (I think) to Crawford's notion of branches (and other
notions of branches). Actually it is the only measure of the search
tree size that makes sense, as more than one variable can be valued
at a single node. 

In particular, a leaf is a solution node or a node (a literal) that
experienced a direct contradiction and whose sibling (the negatation of
the literal) also experienced a direct contradiction. 

*/
extern char* filename;
int LEVEL;
int forcedlits;
int forcedbcls;
int forcedlitsAdded;
int forcedbclsAdded;

int *dplLevelToLit;
bool *dpl2ndBranch;
class ForcedRec;
ForcedRec **dplForcesAtLevel;
ForcedRec **dplOldForcesAtLevel;

double time_out_val;
bool DPL_debug;
bool print_solutions;
int pcr;
bool DOCPB;
bool DOPPB;
bool DOLPB;

extern int *orig_varnum;

void inline addFinalForcedLiteral(int lt,int level, NoGood *ng);
void inline addForcedLiteral(int lt,int level, NoGood *ng);
void inline addForcedBcls(int lt,int x, int level, NoGood *ng);
void inline addFinalForcedBcls(int lt,int x, int level, NoGood *ng);
void inline emptyFatL(int level);
void DoPruneBacks(NoGood *ng, int lt, int btlevel);
void DoForces(int level);

int DPL(int level) {
  int i;
  LEVEL=level; //give global access to current level.

  if (DPL_debug)
    printf("==>Level %d\n",level);
  if (this_num_DPL>0 && (this_num_DPL % 256 == 0)) {
    if((GetInternalRunTime() - start_cpu_time) > (time_out_val)) {
      //Check for time out (more portable than setting up a system
      //interrupt).
      printf("c Timed Out. Fraction of Space explored: ");
      for(i=0;i<level;i++) {
	if(dpl2ndBranch[i])
	  printf("1/2^%d +", i);
      }
      timed_out = 1;
      return(0);
    }
  }
  if(this_num_DPL && pcr && this_num_DPL % pcr == 0) {
    printf("+%d/%d+",HIGHESTBTLEVEL,DEEPESTLEVEL);
    HIGHESTBTLEVEL=level;
    DEEPESTLEVEL=level;
    fflush(stdout);
  }

  //Find next choice point.
  int l;
  int btlevel;
  bool haveChoice=false;
  NoGood *ngfirst=NULL;
  NoGood *ngsecond=NULL;

  while(!haveChoice) {
    l=findNext(level);
    //some findnext functions discover forced literals/bcls. Propagate them.
    propagateStacks();
    //the propagate might force the selected literal.
    //So check if we are finished.
    if(haveContradiction() || l==NOLIT || isActive(l))
      haveChoice=true;
  }
  
  //Check for contradiction.
  if(haveContradiction()) {
    MRFailedLit=l;
    this_num_DPL_leaves++;
    if(!noGoodAtLevel[level])
      noGoodAtLevel[level]=allocNoGood(level);
    noGoodAtLevel[level]->empty();
    noGoodAtLevel[level]->unionFromDAG(&CONTRADICTION.dagnode);

    if(DPL_debug) {
      printf(" DPL found contradiction in findnext\n");
      printConflict(contradictionDAG());
      printf("\n");
      printf(" NoGood Computed\n");
      noGoodAtLevel[level]->print();
    }
    
    btlevel=noGoodAtLevel[level]->max();
    if(!noGoodAtLevel[btlevel])
      noGoodAtLevel[btlevel]=allocNoGood(btlevel);
    noGoodAtLevel[btlevel]->unionFromNoGood(noGoodAtLevel[level]);
    if (DPL_debug) {
      printf("<==Level %d (to %d)\n", level, btlevel);
      printf(" NoGood at %d:            ", level);
      noGoodAtLevel[level]->print();
      printf(" NoGood set at btlevel %d ", btlevel);
      noGoodAtLevel[btlevel]->print();
    }
    if(DOCPB) {
      //DoPruneBacks takes a nogood for a value. Thus
      //we have to remove btlevel from it first
      noGoodAtLevel[level]->unsetbit(btlevel);
      DoPruneBacks(noGoodAtLevel[level],dplLevelToLit[btlevel],btlevel);
    }
    noGoodAtLevel[level]->empty();
    emptyFatL(level);

    if(btlevel<HIGHESTBTLEVEL)
      HIGHESTBTLEVEL=btlevel;
    
    return(btlevel);
  }
    
  if(l==NOLIT) {
    this_num_solutions = 1;
    this_num_DPL_leaves++;
    CheckSolution();
    printf("c Solution Found. Fraction of Space Explored: ");
    for(i=0;i<level;i++) {
      if(dpl2ndBranch[i])
		printf("1/2^%d +", i);
    }
    printf("\n");
    printProblemStats(filename);
    printf("s SATISFIABLE\n");
	
    if (print_solutions) { 
      printf("v ");
      for(int i = 0; i < NLITS; i++) 
	if(isTrue(eqlit(i))) 
	  printf("%d ", to_ff(orig_varnum[variable_of(i)], truth_of(i)));
	  
      printf("0\n");
	  //	printf("Var[%d]=%c (==Var[%d]) at Level %d\n",to_ff(i),
	  //	   ff_tval(val(eqlit(i))),to_ff(eqlit(i)),
	  //	   intTolit(eqlit(i))->assignLevel); 
	}
    //return(true);
    exit(10);
  }
	

  if (level>DEEPESTLEVEL) DEEPESTLEVEL=level;
   
  if (DPL_debug)
    printf("==>Level %d, lit %d   [%d,%d/%d,%d].\n",level,to_ff(l),
	   intTolit(l)->numBcls,intTolit(l)->numNcls,
	   intTolit(negate(l))->numBcls,intTolit(negate(l))->numNcls);
  
  this_num_DPL++;
  dpl2ndBranch[level] = false;
  dplLevelToLit[level]=l;
  markundoCheckPoint(); 

  pushupstack(l,allocDAGNODE_CHOICE(l,level));
  propagateStacks();
  bool firstValueFailed = haveContradiction();

  if (firstValueFailed) {
    ngfirst=allocNoGood(level);
    ngfirst->unionFromDAG(&CONTRADICTION.dagnode);
	
    if(DPL_debug) {
      printf("   1st contradiction at level %d\n",level);
      ngfirst->print();
    }
  }
  else {   //recurse
    if (DPL_debug)
      printf("   DPL Finished propagating first value.\n");
    btlevel = DPL(level + 1);
    if(btlevel<level) {
      DPL_branches_saved++;
      //next two statements should in fact be unnecessary.
      if(noGoodAtLevel[level]) noGoodAtLevel[level]->empty();
      emptyFatL(level);
      if(DPL_debug) 
	printf("<++Level %d direct to %d\n", level, btlevel);
      undoRestore();
      return(btlevel);
    }
  }
  //reset global since we are back at level;
  LEVEL=level;
  //completed recursion now unwind and try second value if necessary. 
  undoRestore();

  if (DPL_debug)
    printf("   DPL About to propagate 2nd branch lit %d at level %d.\n",
	   to_ff(negate(l)),level);

  this_num_DPL++;
  dpl2ndBranch[level] = true;
  dplLevelToLit[level]=negate(l);
  
  markundoCheckPoint(); //Check point
  pushupstack(negate(l),allocDAGNODE_CHOICE(negate(l),level));
  //should in fact only have forces to do if we had recursed.
  //i.e., if ngfirst is empty.
  DoForces(level);
  propagateStacks();
  bool secondValueFailed = haveContradiction();
  if (secondValueFailed) {
    ngsecond=allocNoGood(level);
    ngsecond->unionFromDAG(contradictionDAG());

    if(DPL_debug) {
      printf("   2nd contradiction at level %d\n",level);
      ngsecond->print();
    }
  }
  else { //recurse
    if(DPL_debug)
      printf("   DPL Finished propagating second value.\n");
    btlevel = DPL(level + 1);
    if(btlevel<level) {
      DPL_branches_saved++;
      if(DOPPB) {
	if(!noGoodAtLevel[level])
	  noGoodAtLevel[level]=allocNoGood(level);
	if(ngfirst)
	  noGoodAtLevel[level]->unionFromNoGood(ngfirst);
	DoPruneBacks(noGoodAtLevel[level],l,btlevel+1);
      }
      if(noGoodAtLevel[level]) 
	noGoodAtLevel[level]->empty();
      emptyFatL(level);
      if(ngfirst)
	deallocNoGood(ngfirst);
      if(DPL_debug)
	printf("<==Level %d (to %d)\n", level,btlevel);
      undoRestore();
      return(btlevel);
    }
  }
  
  if (firstValueFailed && secondValueFailed) {
    this_num_DPL_leaves++;
    MRFailedLit=l;
  }

  LEVEL=level;
	
  //if first or second value recursed, and returned to this level.
  //i.e., the recursive call did not backjump above this level.
  //then noGoodAtLevel would have been set. However, we still need
  //to union in ngfirst and ngsecond, if they are non-null.
  //note we did not recurse on either side, noGoodAtLevel[level]
  //will be null.
	
  if(!noGoodAtLevel[level])
    noGoodAtLevel[level]=allocNoGood(level);

  if(ngfirst)
    noGoodAtLevel[level]->unionFromNoGood(ngfirst);
  if(ngsecond)
    noGoodAtLevel[level]->unionFromNoGood(ngsecond);
  
  btlevel=noGoodAtLevel[level]->max();
  if(!noGoodAtLevel[btlevel])
    noGoodAtLevel[btlevel]=allocNoGood(btlevel);
  noGoodAtLevel[btlevel]->unionFromNoGood(noGoodAtLevel[level]);

  //now we have three no-goods to work with.
  
  if (DPL_debug) {
    printf("<==Level %d (to %d)\n", level, btlevel);
    printf(" NoGood at %d:            ", level);
    noGoodAtLevel[level]->print();
    printf(" NoGood set at btlevel %d ", btlevel);
    noGoodAtLevel[btlevel]->print();
  }


  //1. the new nogood passed back to btlevel.
  //   use this to prune back the assignment at btlevel.
  if(DOCPB) {
    //DoPruneBacks takes a nogood for a value. Thus
    //we have to remove btlevel from it first
    noGoodAtLevel[level]->unsetbit(btlevel);
    DoPruneBacks(noGoodAtLevel[level],dplLevelToLit[btlevel],btlevel);
    //
    //2. ngfirst if it exists. It might allow a pruning of l (the lit
    //   on the first branch) to levels above btlevel. Note that
    //   when pruning back the lit at btlevel, we can start by looking
    //   at elements of the nogood less than btlevel, since btlevel
    //   is the max.
    //   However, for ngfirst or ngsecond, the current level is the max.
    //   Hack---we are about to backtrack to btlevel. There is no point
    //   in pruning l or notl to levels between btlevel and level.
    //   so pass doprunebacks btlevel+1 to make it start looking for
    //   a prune back from btlevel and up.
    
    if(ngfirst)
      DoPruneBacks(ngfirst,l,btlevel+1);
    
    //3. similarly with ngsecond
    if(ngsecond)
      DoPruneBacks(ngsecond,negate(l),btlevel+1);
  }
  
  //Note that if either of these do not exist, then the prune back of
  //l or negate l would have been accomplished by some recursive level
  //below for which case 1 corresponds to a prune back of l or negate(l).

  noGoodAtLevel[level]->empty();
  if(ngfirst) deallocNoGood(ngfirst);
  if(ngsecond) deallocNoGood(ngsecond);
  emptyFatL(level);
  undoRestore();  

  if(btlevel<HIGHESTBTLEVEL)
    HIGHESTBTLEVEL=btlevel;

  return(btlevel);
}

void DoPruneBacks(NoGood *ng, int lt, int btlevel) {
  //ng is a value no-good for literal lt. ng is stored as a bit vector
  //in terms levels---the set of assignments at those levels implies
  //that lt cannot hold. We want to compute prunebacks that hold above
  //btlevel
  int i;
  bool pbpossible;
  bool first;
  int notlt=negate(lt);
  //1. Check if a literal prune back is possible
  //printf("=>DoPruneBacks(,%d,%d)\n",to_ff(lt), max);
  //  printf("  Nogood:\n");
  //ng->print();

  pbpossible=false;
  int max=ng->max();
  for(i=max+1;i<btlevel;i++)
    if(!dpl2ndBranch[i]) {
      pbpossible=true;
      break;
    }

  if(pbpossible) {
    first=true;
    //1. Create the nogood
    //   note the nogood must store the index max
    //   so it must be of size nextmax+1.
    NoGood *reducedng=allocNoGood(max+1);
    reducedng->copyNoGood(ng);
    //printf(" DoPruneBack: reducedng:\n");
    //reducedng->print();
    //2. put it on the literal force list for every level
    //   that still has another branch to go.
    for(i=max+1;i<btlevel;i++) {
      if(!dpl2ndBranch[i]) {
	forcedlitsAdded++;
	//printf("DoPruneBack: forcing lit %d at level %d.\n",
	//	   to_ff(notlt), i);
	if(first) {
	  //3. the highest forced literal must deallocate
	  //   the nogood.
	  addFinalForcedLiteral(notlt,i,reducedng);
	  first=false;
	}
	else 
	  addForcedLiteral(notlt,i,reducedng);
      }
    }
  }

  //now look for a binary clause prune back.
  pbpossible=false;
  int nextmax=ng->maxfrom(max);
  for(i=nextmax+1;i<max;i++)
    if(!dpl2ndBranch[i]) {
      pbpossible=true;
      break;
    }
  
  if(pbpossible) {
    first=true;
    NoGood *reducedng=allocNoGood(nextmax+1);
    reducedng->copyNoGood(ng);
    //Find the other lit that forms the binary clause
    int notx=negate(dplLevelToLit[max]);
    for(i=nextmax+1;i<max;i++) {
      if(!dpl2ndBranch[i]) {
	forcedbclsAdded++;
	//printf("DoPruneBack: forcing binary clause (%d,%d) at level %d.\n",
	//     to_ff(notlt), to_ff(notx), i);
	if(first) {
	  addFinalForcedBcls(notlt,notx,i,reducedng);
	  first=false;
	}
	else
	  addForcedBcls(notlt,notx,i,reducedng);
      }
    }
  }

}

class ForcedRec {
  //structure to hold the information about a forced lit or forced bcls.
public:
  int lit1;
  int lit2;
  bool deallocNG;
  //built in space for the DagNode. We can't allocate the Dagnode
  //from allocDAGNODE, as these node are automatically reclaimed
  //on backtrack. We are going to store the forced record at a level the
  //current level.
  DAG Dagnode; 
  ForcedRec *pnext;
  ForcedRec() {
    Dagnode.type=NOGOOD;
    Dagnode.pnext=NULL;
    Dagnode.one=NULL;
    Dagnode.two=NULL;
    deallocNG=false;
  };
};

ForcedRec *pFRalloc;
const int FORCEDRECALLOCFACTOR=200;

void MakeForcedRecStore() {
  ForcedRec *pfr = new ForcedRec[FORCEDRECALLOCFACTOR];
  if(!pfr) 
	panic("Memory Allocation Failure #13\n");
  pFRalloc = pfr;
  //now link the new array.
  for(int i=0;i<FORCEDRECALLOCFACTOR;i++) {
    pfr->pnext=pfr+1;
    pfr++;
  }
  //fix the last pointer
  (--pfr)->pnext=NULL;
}

inline ForcedRec* allocForcedRec() {
  //allocate a new no good that can store levels sz bits inclusive.
  //allocate a new bcls record
  if(!pFRalloc)
    MakeForcedRecStore();
  ForcedRec *pfr = pFRalloc;
  pFRalloc=pfr->pnext;
  return(pfr);
}

inline void deallocForcedRec(ForcedRec *pfr) {
  //deallocate it.
  pfr->pnext = pFRalloc;
  pFRalloc = pfr;
}

inline bool FatLIsEmpty(int level) {
  return(dplForcesAtLevel[level]==NULL);
}

inline void pushFatL(ForcedRec *fr, int level) {
  fr->pnext=dplForcesAtLevel[level];
  dplForcesAtLevel[level]=fr;
} 

inline ForcedRec *popFatL(int level) {
  ForcedRec *fr=dplForcesAtLevel[level];
  dplForcesAtLevel[level]=fr->pnext;
  return(fr);
}

inline bool OldFatLIsEmpty(int level) {
  return(dplOldForcesAtLevel[level]==NULL);
}

inline void pushOldFatL(ForcedRec *fr, int level) {
  fr->pnext=dplOldForcesAtLevel[level];
  dplOldForcesAtLevel[level]=fr;
} 

inline ForcedRec *popOldFatL(int level) {
  ForcedRec *fr=dplOldForcesAtLevel[level];
  dplOldForcesAtLevel[level]=fr->pnext;
  return(fr);
}

inline void emptyFatL(int level) {
  while(!FatLIsEmpty(level))
    deallocForcedRec(popFatL(level));
  while(!OldFatLIsEmpty(level)) {
    ForcedRec *fr=popOldFatL(level);
    if(fr->deallocNG)
      deallocNoGood((NoGood *) fr->Dagnode.one);
    deallocForcedRec(fr);
  }
}

inline void addFinalForcedLiteral(int lt,int level, NoGood *ng) {
  //put lt on the force list at level with a special designation that
  //when this lt is forced we should reclaim ng.
  ForcedRec *fr=allocForcedRec();
  fr->Dagnode.one=(DAG *) ng;
  fr->deallocNG=true;
  fr->lit1=lt;
  fr->lit2=NOLIT;
  pushFatL(fr,level);
}

inline void addForcedLiteral(int lt,int level, NoGood *ng) {
  //identical but set deallocNG to false.
  ForcedRec *fr=allocForcedRec();
  fr->Dagnode.one=(DAG *) ng;
  fr->deallocNG=false;
  fr->lit1=lt;
  fr->lit2=NOLIT;
  pushFatL(fr,level);
}

inline void addFinalForcedBcls(int lt, int x, int level, NoGood *ng) {
  //put lt on the force list at level with a special designation that
  //when this lt is forced we should reclaim ng.
  ForcedRec *fr=allocForcedRec();
  fr->Dagnode.one=(DAG *) ng;
  fr->deallocNG=true;
  fr->lit1=lt;
  fr->lit2=x;
  pushFatL(fr,level);
}

inline void addForcedBcls(int lt,int x, int level, NoGood *ng) {
  //identical but set deallocNG to false.
  ForcedRec *fr=allocForcedRec();
  fr->Dagnode.one=(DAG *) ng;
  fr->deallocNG=false;
  fr->lit1=lt;
  fr->lit2=x;
  pushFatL(fr,level);
}

void DoForces(int level) {
  //process all of the forced bcls and lits
  //Note that we cannot deallocate the ForcedRec's once we force them.
  //...they have data that is required by the search below.
  //Instead we must clear the dplForcesAtLevel and move all of the
  //records over dplOldForcesAtLevel.
  //When we backtrack from this level we can deallocate both the
  //old and the unused new ForcedRecs.

  while(!FatLIsEmpty(level)) {
    ForcedRec *fr=popFatL(level);
    if(fr->lit2==NOLIT) {
      if(!isTrue(fr->lit1)) {
		forcedlits++;
		//printf("DoForces: At Level %d, forcing lit %d\n", level, to_ff(fr->lit1));
		//printf("Nogood:\n");
		//((NoGood *) fr->Dagnode.one)->print();
		pushupstack(fr->lit1,&(fr->Dagnode));
      }
    }
    else {
      if(!haveBclause(fr->lit1,fr->lit2)) {
	pushbclsstack(fr->lit1,fr->lit2,&(fr->Dagnode));
	forcedbcls++;
	//printf("DoForces: At Level %d, forcing bcls (%d,%d)\n", level,
	//     to_ff(fr->lit1), to_ff(fr->lit2));
      }
    }
    pushOldFatL(fr,level);
  }
}

void initDPL() {
  dplLevelToLit = new int[NACTIVELITS/2];
  dpl2ndBranch = new bool[NACTIVELITS/2];
  dplForcesAtLevel = new (ForcedRec *[NACTIVELITS/2]);
  dplOldForcesAtLevel = new (ForcedRec *[NACTIVELITS/2]);

  if(!dplLevelToLit || !dpl2ndBranch || !dplForcesAtLevel
	 || !dplOldForcesAtLevel)
	panic("Memory Allocation Failure #14\n");

  for(int i=0;i<NACTIVELITS/2;i++) {
    dplLevelToLit[i] = NOLIT;
    dpl2ndBranch[i] = false;
    dplForcesAtLevel[i] = NULL;
    dplOldForcesAtLevel[i] = NULL;
  }
  forcedlits=0;
  forcedbcls=0;
  forcedlitsAdded=0;
  forcedbclsAdded=0;
  pFRalloc=NULL;
}
